L_HGmodify = "L_HGmodify"
L_HGdelete = "L_HGdelete"
L_FAdelete = "L_FAdelete"
L_FAmodify = "L_FAmodify"
L_FAether = "L_FAether"
L_merge = "L_merge"
LCB_generation = "LCB_generation"
FFA_modify = "FFA_modify"
FFA_elong = "FFA_elong"
FFA_oxi = "FFA_oxi"
FFA_sat = "FFA_sat"
FFA_desat = "FFA_desat"
FFA_delete = "FFA_delete"
